%==========================================================================
% Run NIMBLE on Caltech-101. See the demo code for pedagogical purposes.
% This version should perform similarly to the results reported in
% Kanan & Cottrell (2010). You should get something like this:
% ----------------------------
% |  Train  | Mean Per Class |
% |  Number |   Accuracy     |
% |---------------------------
% |    5    |      55.9%     |
% |   15    |      70.8%     |
% ----------------------------
%
%==========================================================================
% Author:   Christopher Kanan (chriskanan@gmail.com)
%
% Cite as:  Kanan, C. & Cottrell, G. W. (2010). Robust classification of
%             objects, faces, and flowers using natural image Statistics.
%             In Proceedings of the IEEE Conference on Computer Vision
%             and Pattern Recognition (CVPR), 2010.
%==========================================================================

clear;


%set this to your Caltech-101 directory
base_dir = fullfile('/home/ckanan/Datasets', 'Caltech 101');
%base_dir = ''; 

%times to fixate
max_train_samples = 100;
max_test_samples = 100;
%max_train_samples = 30;
%max_test_samples = 30;
num_validation_runs = 1;

verbose = true;

[ret, computer_name] = system('hostname');
computer_name = lower(computer_name(1:end-1));

%number of training instances and test instances
max_class = inf;
%max_class = 5;
%train_list = [15, 30, 5, 1];
train_list = 5; %num train
train_list = 15; %num train
num_test = 20; %num test        

%many parameters                
whiten = true;        
max_principal_components = 500;
%max_principal_components = 300;
Model.img_size = 128;

%set up model parameters
Model.grid_list = [1, 2, 4];
Model.use_probit = false;
Model.pooling_method = 1;
Model.parallel_mode = true;
fftw('planner', 'measure');


%Load up SUN
SUN = load('SUN.mat');
SUN = SUN.SUN;

try
    SUN = rmfield(SUN, 'Separating_Matrix');
    SUN = rmfield(SUN, 'basis_vectors');
end
Model.SUN = SUN;
clear SUN;

dataset_name = 'Caltech 101';

%reset_rand is used to ensure that the results are identical on a 
%particular machine per run, if you want to tune parameters and process
%images in parallel
%Model.reset_rand = false;
Model.reset_rand = true;
Model.retina_radius = floor(Model.img_size / 5);
Model.SUN.pad_image = true; 

fprintf('Running experiment on %s!\n', dataset_name);
fprintf('retina_radius = %d    filt_size = %d    filter_num = %d\n', Model.retina_radius, size(Model.SUN.Filters(1).R, 1), length(Model.SUN.Filters));
fprintf('Using %d Train Samples and %d Test Samples.\n', max_train_samples, max_test_samples);

big_iter = 0;
Results.max_accuracy = cell(length(train_list), 1);
Results.final_accuracy = cell(length(train_list), 1);
Results.mean_auc = cell(length(train_list), 1);
Results.mean_class_accuracy = cell(length(train_list), 1);
Results.temporal_accuracy = cell(length(train_list), 1);
Results.max_principal_components = max_principal_components;
Results.train_list = train_list;
Results.num_validation_runs = num_validation_runs;
Results.retina_radius = Model.retina_radius;
Results.num_test = num_test;

total_iters = length(train_list)*num_validation_runs;
total_time = 0;
timer_iter = 0;

for num_train = train_list
    big_iter = big_iter + 1;

    [train_files, test_files, dataset_ndx, Class_Names] = Load_Caltech101(num_train, num_test, max_class, base_dir);
    
    Results.max_accuracy{big_iter} = zeros(num_validation_runs, 1);
    Results.final_accuracy{big_iter} = zeros(num_validation_runs, 1);
    Results.mean_auc{big_iter} = zeros(num_validation_runs, 1);
    Results.mean_class_accuracy{big_iter} = zeros(num_validation_runs, 1);
    Results.temporal_accuracy{big_iter} = zeros(num_validation_runs, max_test_samples);
    
    %cross validation runs
    for v = 1:num_validation_runs        
        my_time = tic();                 
        timer_iter = timer_iter + 1;
        if verbose, fprintf('Starting run %d of %d...\n', v, num_validation_runs);end;
        
        %ensure that these parameters are uninitialized
        Model.PCA_Mean = [];
        Model.PCA_Vectors = [];
        
        %load the data
        if Model.reset_rand
            RandStream.setDefaultStream(RandStream('mt19937ar','seed',big_iter*v*timer_iter));
        end
        [train_files, test_files, dataset_ndx, Class_Names] = Load_Caltech101(num_train, num_test, max_class, base_dir);
        
        if v == 1 && verbose, fprintf('%d Classes total: Using %d train and %d test images per class.\n', size(train_files, 1), size(train_files, 2), size(test_files, 2));end
        
        %acquire training data
        if verbose, fprintf('Acquiring training data...\n');end;
        [Model, Stats] = Train_And_Test_Model(Model, train_files, 1, max_train_samples, Class_Names, Model.img_size, verbose);

        
        train_files = [];
        
        try
            %this is done to free a considerable amount of memory,
            %especially on Linux machines
            matlabpool close force;
        end
        
        %Do PCA on acquired data
        if verbose, fprintf('Performing PCA...\n');end;       
        train_data = Stats.training_data;
        train_labels = Stats.training_labels;
        train_locations = Stats.training_locations;        
        Stats = [];
                
        %Do Whitened PCA
        %[Model.PCA_Mean, Model.PCA_Vectors, train_data] = Whitened_PCA(train_data, max_principal_components, false);       
        
        train_data(~isfinite(train_data)) = 0;
      
        Model.PCA_Mean = mean(train_data, 2);
        
        %subtract mean
        train_data = bsxfun(@minus, train_data, Model.PCA_Mean);        
        
        %pca
        train_data = train_data';
        [U,sigma,PCA_Vectors] = pca(train_data, max_principal_components);        
        train_data = train_data';        
        clear U;
        PCA_Vectors = PCA_Vectors';
        max_principal_components = min(max_principal_components, size(PCA_Vectors, 1));
        PCA_Vectors = PCA_Vectors(1:max_principal_components, :);       
        PCA_Vectors(~isreal(PCA_Vectors)) = 0;
        
        %whitening        
        pca_train_data = PCA_Vectors * train_data;
        Whitening_Mat = (pca_train_data * pca_train_data');
        Whitening_Mat(Whitening_Mat < 0) = 0;
        Whitening_Mat = Whitening_Mat^(-1/2);
        Whitening_Mat(~isreal(Whitening_Mat)) = 0;
        PCA_Vectors = Whitening_Mat * PCA_Vectors;
        clear pca_train_data
        clear Whitening_Mat
                
        train_data = PCA_Vectors * train_data;
        
        Model.PCA_Vectors = PCA_Vectors;
        clear PCA_Vectors;
        
        
        %train the classifier: 1-NN kernel density estimator
        if verbose, fprintf('Training classifier...\n');end;
        norm_data = sqrt(sum(train_data .^ 2));
        norm_data(norm_data == 0) = 1;
        if max(abs(norm_data - 1)) > 1e-5
            train_data = bsxfun(@rdivide, train_data, norm_data);
        end
        
        Model.Net = NN_KDE_Train(train_data, train_labels);
        Model.Net.W_Locations = train_locations;
        train_data = [];
       
        %evaluate model
        if verbose, fprintf('Evaluating performance...\n');end;
        [Model, Test_Stats] = Train_And_Test_Model(Model, test_files, 2, max_test_samples, Class_Names, Model.img_size, verbose);
        
        
        %save stats
        Results.max_accuracy{big_iter}(v) = Test_Stats.Max_Accuracy;
        Results.final_accuracy{big_iter}(v) = Test_Stats.Final_Accuracy;
        Results.mean_auc{big_iter}(v) = Test_Stats.Mean_AUC; %this is not ROC, it is the area under the accuracy as a function of fixations
        Results.mean_class_accuracy{big_iter}(v) = Test_Stats.Mean_Class_Accuracy;
        Results.temporal_accuracy{big_iter}(v, :) = Test_Stats.Temporal_Accuracy;
        
        total_time = total_time + ((toc(my_time) / 60) / 60);
        
        fprintf('Estimated Time Remaining: %.2f hours\n', (total_time / timer_iter)*(total_iters - timer_iter));
        
        Model.Net = [];
       
        try
         %   save(sprintf('Results %s by %s.mat', Dataset_List{Dataset_Index}, computer_name), 'Results');
        catch
        end
              
        pause(.01);
    end
           
    %calculate summary statistics    
    %n = sqrt(size(test_files, 1)); %for Standard error of the mean
    n = 1; %plain std
    fprintf('Max Accuracy: %.4f  %.4f\n', mean(Results.max_accuracy{big_iter}), std(Results.max_accuracy{big_iter}) / n);
    fprintf('Final Accuracy: %.4f  %.4f\n', mean(Results.final_accuracy{big_iter}), std(Results.final_accuracy{big_iter}) / n);
    fprintf('Mean Final Class Accuracy: %.4f  %.4f\n', mean(Results.mean_class_accuracy{big_iter}), std(Results.mean_class_accuracy{big_iter}) / n);
            
    pause(.01);    
end
